<?php

namespace Drupal\calendar_view\Plugin\views\style;

use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\datetime\Plugin\Field\FieldType\DateTimeItem;
use Drupal\datetime\Plugin\Field\FieldType\DateTimeItemInterface;
use Drupal\datetime_range\Plugin\Field\FieldType\DateRangeItem;
use Drupal\views\Plugin\views\filter\Date;
use Drupal\views\Plugin\views\style\DefaultStyle;
use Drupal\views\ResultRow;

/**
 * Unformatted style plugin to render rows one after another with no
 * decorations.
 *
 * @ingroup views_style_plugins
 *
 * @ViewsStyle(
 *   id = "calendar",
 *   title = @Translation("Calendar"),
 *   help = @Translation("Displays rows in a calendar."),
 *   theme = "views_view_calendar",
 *   display_types = {"normal"}
 * )
 */
class Calendar extends DefaultStyle {

  /**
   * Retrieve all Date fields exposed as filters.
   *
   * @return array
   *   List of exposed filter names, keyed by field ID.
   */
  protected function getDateFilters() {
    $date_filters = [];
    $filters = $this->view->display_handler->getHandlers('filter');
    foreach ($filters as $filter_id => $filter) {
      if ($filter instanceof Date && $info = $filter->exposedInfo()) {
        $date_filters[$filter_id] = $info['label'] ?? $filter_id;
      }
    }
    return $date_filters;
  }

  /**
   * Get default options, statically.
   *
   * @return array
   */
  public static function getDefaultOptions() {
    return [
      'calendar_filters' => [],
      'calendar_type' => 'monthly',
      'calendar_weekday_start' => 1, // Monday
    ];
  }

  /**
   * {@inheritdoc}
   */
  protected function defineOptions() {
    $options = parent::defineOptions();

    $defaults = self::getDefaultOptions();
    foreach ($defaults as $key => $value) {
      $options[$key] = ['default' => $value];
    }

    return $options;
  }

  /**
   * {@inheritdoc}
   */
  public function buildOptionsForm(&$form, FormStateInterface $form_state) {
    parent::buildOptionsForm($form, $form_state);

    $date_filters = $this->getDateFilters();
    $date_filter_keys = array_keys($date_filters);
    $default_date_filter = [reset($date_filter_keys)];

    $form['calendar_filters'] = [
      '#type' => 'checkboxes',
      '#title' => $this->t('Date fields'),
      '#empty_option' => $this->t('- Select -'),
      '#options' => $date_filters,
      '#default_value' => $this->options['calendar_filters'] ?? $default_date_filters,
      '#disabled' => empty($date_filters),
    ];
    if (empty($date_filters)) {
      $form['calendar_filters']['#description'] = $this->t('Empty? You need to add a date field in exposed filters on this View.');
    }

    $form['calendar_weekday_start'] = [
      '#type' => 'select',
      '#title' => $this->t('Start week on:'),
      '#options' => [
        1 => t('Monday'),
        2 => t('Tuesday'),
        3 => t('Wednesday'),
        4 => t('Thursday'),
        5 => t('Friday'),
        6 => t('Saturday'),
        0 => t('Sunday'),
      ],
      '#default_value' => $this->options['calendar_weekday_start'] ?? 1,
    ];

    // @todo Work in progress.
    $form['calendar_type'] = [
      '#type' => 'select',
      '#title' => $this->t('Calendar type'),
      '#options' => [
        'monthly' => $this->t('Monthly'),
        'weekly' => $this->t('Weekly'),
        'agenda' => $this->t('Agenda'),
      ],
      '#default_value' => $this->options['calendar_type'] ?? 'monthly',
      '#disabled' => TRUE,
      '#wrapper_attributes' => ['class' => ['visually-hidden']],
    ];
  }

  /**
   *
   */
  public static function processResults(string $field_name, array $results) {
    $content = [];
    foreach ($results as $result) {
      $view_row = $result['content']['#row'] ?? NULL;
      $entity = $view_row instanceof ResultRow ? $view_row->_entity : NULL;

      try {
        if (!$entity instanceof EntityInterface) {
          throw new \Exception('Entity not found.');
        }
        if (!$entity->hasField($field_name)) {
          throw new \Exception('Field does not exist: ' . $field_name);
        }
        if (($items = $entity->get($field_name))->isEmpty()) {
          throw new \Exception('Field is empty: ' . $field_name);
        }
        if (!$items) {
          throw new \Exception('Field items undefined.');
          continue;
        }

        foreach (range(1, $items->count()) as $i) {
          $item = $items->get($i - 1);

          // Default date field values.
          $timestamp = $item->value;
          $result['datetime'] = DateTimeItem::DATETIME_TYPE_DATETIME;
          $result['from'] = $result['to'] = $timestamp;
          $result['calendar_filter'] = $field_name;

          // Datetime field.
          if ($item instanceof DateTimeItemInterface) {
            $datetime_type = $item->getFieldDefinition()->getSetting('datetime_type');
            $result['datetime'] = $datetime_type;

            // Get date(time) or range.
            if ($item instanceof DateRangeItem) {
              $result['from'] = $item->get('value')->getDateTime()->getTimestamp();
              $result['to'] = $item->get('end_value')->getDateTime()->getTimestamp();
            } elseif ($item instanceof DateTimeItem) {
              $result['from'] = $result['to'] = $item->date->getTimestamp();
            }
          }
          // Insert content occurences.
          if ($from = $result['from'] ?? NULL) {
            $time_date = date('Y-m-d 00:00:00', $from);
            $timestamp = (int) date('U', strtotime($time_date));
            $content[$timestamp][] = $result;
          }
          if ($to = $result['to'] ?? NULL) {
            $date_time_from = new \DateTime();
            $date_time_from->setTimestamp($from);

            $date_time_to = new \DateTime();
            $date_time_to->setTimestamp($to);

            $interval = $date_time_to->diff($date_time_from);

            // Calculcate days span.
            $is_same_day = $interval->d == 0;
            $is_all_day = DateRangeItem::DATETIME_TYPE_ALLDAY;
            $start = $is_same_day && $is_all_day ? 1 : 0;
            for ($i = $start; $interval->d >= $i; $i++) {
              $date_time_from->modify('+1 day');
              $time_date = date('Y-m-d 00:00:00', $date_time_from->format('U'));
              $timestamp = (int) date('U', strtotime($time_date));
              $content[$timestamp][] = $result;
            }
          }
        }
      } catch (\Exception$e) {
        // Fail silently.
        $content['errors'][] = $result;
      }
    }

    // Sort content by timestamp.
    asort($content);

    return $content;
  }

  /**
   * Render a calendar.
   *
   * @todo How to start on another day?
   * @todo Make weekday label configurable.
   * @todo How to trim filter field name correctly?
   */
  public static function buildCalendar(array $options = [], array $results = []) {
    // Prepare view results to insert in table.
    $calendar_filters = $options['calendar_filters'] ?? [];
    $calendar_filters = array_filter($calendar_filters, function ($value) {
      return $value !== 0;
    });
    if (!$calendar_filters) {
      return;
    }

    $contents = [];
    foreach ($calendar_filters as $calendar_filter) {
      // Ugly... how to get field name correctly?
      $field_name = str_replace('_value', '', $calendar_filter);

      // Process results for calendar display.
      $list = self::processResults($field_name, $results);
      foreach ($list as $timestamp => $rows) {
        if (!isset($contents[$timestamp])) {
          $contents[$timestamp] = [];
        }
        $contents[$timestamp] += $rows;
      }
    }

    // Render an empty calendar if no results.
    $timestamps = array_filter(array_keys($contents), function ($value) {
      return ctype_digit($value);
    });
    if (empty($timestamps)) {
      $default_time = $options['context']['timestamp'] ?? date('U');
      $timestamps = [$default_time];
    }
    // Build calendars.
    $calendars = [];
    foreach ($timestamps as $timestamp) {
      $year = date('Y', $timestamp);
      $month = date('m', $timestamp);
      if (!isset($calendars[$year . $month])) {
        $calendars[$year . $month] = self::buildMonthTable($options, $year, $month);
      }
    }

    // Populate calendars.
    foreach ($calendars as $key => $table) {
      foreach ($table['#rows'] ?? [] as $i => $rows) {
        foreach ($rows['data'] ?? [] as $timestamp => $cells) {
          $cell = &$calendars[$key]['#rows'][$i]['data'][$timestamp]['data'];
          $cell['#children'] = $contents[$timestamp] ?? [];
        }
      }
    }

    return $calendars;
  }

  /**
   * Render a month calendar as a table.
   */
  public static function buildMonthTable($options, $year, $month) {
    // Build table.
    $days = [
      1 => t('Monday'),
      2 => t('Tuesday'),
      3 => t('Wednesday'),
      4 => t('Thursday'),
      5 => t('Friday'),
      6 => t('Saturday'),
      0 => t('Sunday'),
    ];

    $weekday_start = $options['calendar_weekday_start'] ?? 0;
    $weekdays = range($weekday_start, 6);
    $days = array_replace(array_flip($weekdays), $days);

    $headers = [];
    foreach ($days as $number => $name) {
      $headers[$number] = $name;
    }

    // Dates for this month.
    $month_start = strtotime($year . '-' . $month . '-' . '01');
    $month_days = date('t', $month_start);
    $first_day = date('w', $month_start);
    $month_weeks = floor(($first_day + $month_days - 1) / 7);
    $month_end = strtotime($year . '-' . $month . '-' . $month_days);

    // Next month.
    $next_month = $month == '12' ? '01' : str_pad(($month + 1), 2, '0', STR_PAD_LEFT);
    $next_year = $month == '12' ? $year + 1 : $year;

    // Last month.
    $previous_month = $month == '01' ? '12' : str_pad(($month - 1), 2, '0', STR_PAD_LEFT);
    $previous_year = $month == '01' ? $year - 1 : $year;
    $previous_month_start = strtotime($previous_year . '-' . $previous_month . '-' . '01');
    $previous_month_days = date('t', $previous_month_start);

    $previous_month_offset = [];
    $month_started = FALSE;
    foreach (array_keys($headers) as $number) {
      if ($month_started = (int) $number == (int) $first_day) {
        break;
      }
      $previous_month_offset[] = $previous_month_days;
      $previous_month_days--;
    }
    $previous_month_offset = array_reverse($previous_month_offset);

    $count = 0;
    $month_started = FALSE;
    for ($i = 0; $i <= $month_weeks; $i++) {
      // Prepare row.
      $cells = [];

      // First week.
      if ($i == 0) {
        // Empty days starting the month display.
        foreach ($previous_month_offset as $daynum) {
          $day_number = str_pad($daynum, 2, '0', STR_PAD_LEFT);
          $time_now = strtotime($previous_year . '-' . $previous_month . '-' . $day_number);

          $cells[$time_now]['data'] = [
            '#theme' => 'calendar_view_day',
            '#timestamp' => $time_now,
            '#children' => [],
          ];
          $cells[$time_now]['data-calendar-view-day'] = date('d', $time_now);
          $cells[$time_now]['data-calendar-view-month'] = date('m', $time_now);
          $cells[$time_now]['data-calendar-view-year'] = date('y', $time_now);
          $cells[$time_now]['class'] = ['previous-month'];
        }

        // Pending days of this month's first week.
        $x = 7 - count($previous_month_offset);
        do {
          $x--;

          // Count days of the month.
          $count++;

          // Get this day's timestamp.
          $day_number = str_pad($count, 2, '0', STR_PAD_LEFT);
          $time_now = strtotime($year . '-' . $month . '-' . $day_number);

          $cells[$time_now]['data'] = [
            '#theme' => 'calendar_view_day',
            '#timestamp' => $time_now,
            '#children' => [],
          ];
          $cells[$time_now]['data-calendar-view-day'] = date('d', $time_now);
          $cells[$time_now]['data-calendar-view-month'] = date('m', $time_now);
          $cells[$time_now]['data-calendar-view-year'] = date('y', $time_now);
          $cells[$time_now]['class'] = ['current-month'];
        } while ($x >= 1);

        // Populate table row.
        $rows[] = ['data' => $cells];

        continue;
      }

      // Rest of the weeks.
      $daynum = 0;
      foreach (array_keys($headers) as $number) {
        // Count days of the month.
        $count++;

        // Fill next months day, if necessary.
        $month_finished = $count > (int) $month_days;
        $week_finished = $number == count($headers);
        if ($month_finished && !$week_finished) {
          $daynum++;
          $time_now = strtotime($next_year . '-' . $next_month . '-' . $daynum);
          $cells[$time_now]['data'] = [
            '#theme' => 'calendar_view_day',
            '#timestamp' => $time_now,
            '#children' => [],
          ];
          $cells[$time_now]['data-calendar-view-day'] = date('d', $time_now);
          $cells[$time_now]['data-calendar-view-month'] = date('m', $time_now);
          $cells[$time_now]['data-calendar-view-year'] = date('y', $time_now);
          $cells[$time_now]['class'] = ['next-month'];
          continue;
        }

        // Stop now.
        if ($month_finished) {
          break;
        }

        // Insert day.
        $day_number = str_pad($count, 2, '0', STR_PAD_LEFT);
        $time_now = strtotime($year . '-' . $month . '-' . $day_number);

        $cells[$time_now]['data'] = [
          '#theme' => 'calendar_view_day',
          '#timestamp' => $time_now,
          '#children' => [],
        ];
        $cells[$time_now]['data-calendar-view-day'] = date('d', $time_now);
        $cells[$time_now]['data-calendar-view-month'] = date('m', $time_now);
        $cells[$time_now]['data-calendar-view-year'] = date('y', $time_now);
        $cells[$time_now]['class'] = ['current-month'];
      }

      // Populate table row.
      $rows[] = ['data' => $cells];
    }

    return [
      '#type' => 'table',
      '#caption' => date('F Y', $month_start),
      '#header' => $headers,
      '#rows' => $rows,
      '#empty' => NULL,
      '#attributes' => [
        'data-calendar-view-year' => $year,
        'data-calendar-view-month' => $month,
        'class' => [
          'calendar-view-table',
        ],
      ],
    ];
  }
}
